require "thor"
require "uri"
require "net/http"
require "json"
require "table_print"
require "date"

require 'ignition/load_common'

#     version             print version
#     scan                scan for running igntion daemons (UDP broadcast) 
#     send                send a bundle definiion
#     info                print bundle information
#     services            print service information
#     load                load a bundle
#     run                 run a bundle
#     stop                stop a bundle                          


class Ignition::CLI < Thor

  class << self
    def add_shared_option(name, options = {})
      @shared_options = {} if @shared_options.nil?
      @shared_options[name] =  options
    end

    def shared_options(*option_names)
      option_names.each do |option_name|
        opt =  @shared_options[option_name]
        raise "Tried to access shared option '#{option_name}' but it was not previously defined" if opt.nil?
        option option_name, opt
      end
    end

    # Hackery. Take the run method away from Thor so that we can redefine it.
    def is_thor_reserved_word?(word, type)
      return false if word == "run"
      super
    end
  end

  add_shared_option :hostname, type: :string,
                         default: "localhost",
                         desc: "host address to load the file onto."

  add_shared_option :port, type: :numeric,
                           default: 6753,
                           desc: "host port for the ignition daemon"

  map ["-v", "--version"] => :version

  desc "nodes", "scans the local network for conneted ignition daemons"
  shared_options :hostname, :port
  def nodes
    # puts "Run Scan #{options[:port]}"
    # puts "TODO"

    uri = URI.parse("http://#{options[:hostname]}:#{options[:port]}/v1/ignition/nodes")
    http = Net::HTTP.new(uri.host, uri.port)
    request = Net::HTTP::Get.new(uri.request_uri)
    response = http.request(request)

    # puts response.body
    if response.code.to_i == 200
      nodes = JSON.parse(response.body)
      # BundleHelper.print_bundles(options[:hostname],bundles)
      # pp nodes
      tp(nodes.map{ |k,v| {:hostname => k, :last_seen => v, :secs_ago => (DateTime.now.to_time - DateTime.parse(v).to_time).to_i } })
    else
      puts "ERROR - #{response.code}"
      puts response.body
    end
  end

  desc "load [bundle-file]", "load the bundle file into the ignition daemon"
  shared_options :hostname, :port
  def load(bundle_path)
    if (!File.exists?(bundle_path))
      puts "ERROR: bundle file missing on path #{bundle_path}"
    else
      puts "load file:#{bundle_path}"
      puts "load hostname:#{options[:hostname]}"
      puts "load port:#{options[:port]}"

      raw_data = File.read(bundle_path)
      data = JSON.parse(raw_data)
      # puts data.class
      # pp data.to_hash
      params = {'data' => raw_data}

      pp params
      
      x = Net::HTTP.post_form(URI.parse("http://#{options[:hostname]}:#{options[:port]}/v1/ignition/bundles"), params)
      pp x

      puts x.body  
      puts x.code
    end
  end

  desc "run [bundle-id]", "run the bundle id"
  shared_options :hostname, :port
  method_option  :scale,   type: :numeric,
                           default: 1,
                           desc: "scale to X running processes"
  def run(bundle_id)
      puts "run bundle id #{bundle_id}"
      puts "load hostname:#{options[:hostname]}"
      puts "load port:#{options[:port]}"
      puts "load scale:#{options[:scale]}"

      # Net::HTTP.post_form(URI.parse("http://#{options[:hostname]}:#{options[:port]}/v1/ignition/bundles"), params)
      uri = URI.parse("http://#{options[:hostname]}:#{options[:port]}/v1/ignition/bundle/#{bundle_id}/scale/#{options[:scale]}")
      http = Net::HTTP.new(uri.host, uri.port)
      request = Net::HTTP::Get.new(uri.request_uri)
      response = http.request(request)

      puts response.body  
      puts response.code
  end

  desc "info", "info on the bundles"
  shared_options :hostname, :port
  method_option  :bundle_id,   type: :string,
                               default: "*",
                               desc: "info on only this bundle"
  def info
      puts "info bundle id #{options[:bundle_id]}"
      puts "load hostname:#{options[:hostname]}"
      puts "load port:#{options[:port]}"
      puts "load scale:#{options[:scale]}"

      uri = URI.parse("http://#{options[:hostname]}:#{options[:port]}/v1/ignition/bundles")
      http = Net::HTTP.new(uri.host, uri.port)
      request = Net::HTTP::Get.new(uri.request_uri)
      response = http.request(request)

      if response.code.to_i == 200
        bundles = JSON.parse(response.body)
        if options[:bundle_id] != "*"
          bundles.reject!{ |b| b.id != options[:bundle_id]}
        end
        BundleHelper.print_bundles(options[:hostname],bundles)
      else
        puts "ERROR - #{response.code}"
        puts response.body
      end
  end

  desc "stop [bundle-id]", "stop the bundle id"
  shared_options :hostname, :port
  method_option  :scale,   type: :numeric,
                           default: 0,
                           desc: "scale down to X running processes"
  def stop(bundle_id)
      puts "run bundle id #{bundle_id}"
      puts "load hostname:#{options[:hostname]}"
      puts "load port:#{options[:port]}"
      puts "load scale:#{options[:scale]}"

      # Net::HTTP.post_form(URI.parse("http://#{options[:hostname]}:#{options[:port]}/v1/ignition/bundles"), params)
      uri = URI.parse("http://#{options[:hostname]}:#{options[:port]}/v1/ignition/bundle/#{bundle_id}/scale/#{options[:scale]}")
      http = Net::HTTP.new(uri.host, uri.port)
      request = Net::HTTP::Get.new(uri.request_uri)
      response = http.request(request)

      puts response.body  
      puts response.code
  end

  desc "unload", "unload the bundle"
  shared_options :hostname, :port
  method_option  :bundle_id,   type: :string,
                               default: "*",
                               desc: "info on only this bundle"
  def unload(bundle_id)
      puts "info bundle id #{options[:bundle_id]}"
      puts "load hostname:#{options[:hostname]}"
      puts "load port:#{options[:port]}"
      puts "load scale:#{options[:scale]}"

      uri = URI.parse("http://#{options[:hostname]}:#{options[:port]}/v1/ignition/bundle/#{bundle_id}")
      http = Net::HTTP.new(uri.host, uri.port)
      req = Net::HTTP::Delete.new(uri.path)
      response = http.request(req)
      puts "deleted #{response}"

      if response.code.to_i == 200
        puts "all ok"
      else
        puts "ERROR - #{response.code}"
        puts response.body
      end
  end


  desc "logs", "logs for the bundle"
  shared_options :hostname, :port
  method_option  :bundle_id,   type: :string,
                               default: "*",
                               desc: "logs for this bundle id"
  def logs(bundle_id)
      puts "info bundle id #{options[:bundle_id]}"
      puts "load hostname:#{options[:hostname]}"
      puts "load port:#{options[:port]}"

      uri = URI.parse("http://#{options[:hostname]}:#{options[:port]}/v1/ignition/bundle/#{bundle_id}/logs")
      http = Net::HTTP.new(uri.host, uri.port)
      req = Net::HTTP::Get.new(uri.path)
      response = http.request(req)

      if response.code.to_i == 200
        # puts "all ok"

        logs = JSON.parse(response.body)

        # pp logs
        logs.each do |log|
          puts log.strip
        end
      else
        puts "ERROR - #{response.code}"
        puts response.body
      end
  end

end