require "socket"
require "thread"
require "ipaddr"
require "securerandom"

class BroadcastClient
  MULTICAST_ADDR = "224.6.8.11"
  # BIND_ADDR = "0.0.0.0"
  # PORT = 6811

  def initialize(handle,bind_address,bind_port)
    @handle    = handle
    @client_id = SecureRandom.hex(5)
    @listeners = []

    @bind_addr = bind_address
    @port = bind_port

    @stopped = false
  end

  def add_message_listener(listener)
    listen unless listening?
    @listeners << listener
    # puts "add_message_listener 1"
  end

  def transmit(content)
    message = BroadcastMessage.new(
      "client_id" => @client_id,
      "handle"    => @handle,
      "content"   => content
    )

    socket.send(message.to_json, 0, MULTICAST_ADDR, @port)
    message
  end

  def stop!
    # puts "Stopping BroadcastClient"
    @stopped = true
  end

  private

  def listen
    puts "listen"
    socket.bind(@bind_addr, @port)

    Thread.new do
      begin
        while(!@stopped)
          # puts "waiting on socket"
          attributes, _ = socket.recvfrom(1024)
          message = BroadcastMessage.inflate(attributes)
          # puts "got data"
          # unless message.client_id == @client_id
            @listeners.each { |listener| listener.new_message(message) }
          # end
        end
      rescue Exception => exc
        puts exc.message
        puts exc.backtrace
      end
      # puts "closing socket"
      socket.close
    end

    @listening = true
  end

  def listening?
    puts "listening?"
    @listening == true
  end

  def socket
    @socket ||= UDPSocket.open.tap do |socket|
      socket.setsockopt(Socket::IPPROTO_IP, Socket::IP_ADD_MEMBERSHIP, bind_address)
      socket.setsockopt(Socket::IPPROTO_IP, Socket::IP_MULTICAST_TTL, 1)
      socket.setsockopt(Socket::SOL_SOCKET, Socket::SO_REUSEADDR, 1)
    end
  end

  def bind_address
    IPAddr.new(MULTICAST_ADDR).hton + IPAddr.new(@bind_addr).hton
  end
end