require 'pp'
require 'json'

require "foreman/engine"

class Ignition::Manager < Foreman::Engine
	attr_reader :env

	BUNDLES_JSON = "#{Ignition::ROOT_FOLDER}/bundles.json"
	BUNDLES_STATUS_JSON = "#{Ignition::ROOT_FOLDER}/bundles_status.json"

	def initialize(broadcast_client)
		#load the plugs
		super()

		@recent_logs = {}

		@registered_bundles = {}

		@process_ids = {}

		@shutdown = false

		@broadcast_client = broadcast_client

		@broadcast_client.add_message_listener(self)

		@network_daemons = {}

		env = ENV
		
		@bundles = load_bundles

		BundleHelper.prepare_bundles(@bundles)

		register_bundles

		register_signal_handlers

		start_network_ping
	end

	def bundles
		@bundles
	end

	def register_bundles
		@bundles.each do |b|
			register_bundle(b)
		end
		pp @registered_bundles
	end

	def register_bundle(bundle)
		puts "Register bundle #{bundle.id}"

		options = {}
		options[:cwd] = bundle.cwd 
		
		proc = register(bundle.id,bundle.command,options)

		@registered_bundles[proc] = bundle
	end

	# start from foreman engine
	def startup
	end

	def output(name, data)
	  	log = "#{Time.now.strftime("%H:%M:%S.%L")} - #{data}"
	  	puts "#{name} - #{log}"

	  	if @recent_logs[name].nil?
	  		@recent_logs[name] = FixedArray.new(100)
	  	end
	  	@recent_logs[name].push(log)
	end

	def shutdown
	end

	# end from foreman engine

	def load_bundle(bh)
		if (!bundle(bh.id).nil?)
			raise "Bundle Exists. Unload bundle before adding."
		end

		bundle = Ignition::Bundle.new(bh)
		BundleHelper.prepare_bundle(bundle)

		@bundles << bundle
		register_bundle(bundle)
		flush
	end

	def unload_bundle(bh)
		if (bundle(bh.id).nil?)
			raise "Bundle #{bh.id} does NOT exist. Cannot unload."
		end

		#check for running bundles. if there are then raise error. need to be stopped first.

		# bundle = Ignition::Bundle.new(bh)
		# @bundles << bundle

		@bundles.reject! {|b| b.id == bh.id }

		# register_bundle(bundle)
		flush
	end

	# def remove_bundle

	# end

	def run(on_shutdown)
		Thread.new {
			begin
				spawn_processes
			    watch_for_output
			    flush
			    sleep 0.1
			    
			    while(!@shutdown)
			    # loop {
		    	    pid = watch_for_termination { 
			    		puts "child died"
			    		# terminate_gracefully 
			    	}
			    	# puts "got pid notification #{pid}"
					if pid.nil?
						sleep(1)
						# puts "waiting for child process"
					end
				end
				puts "spawn thread complete."
				# loop { sleep(1000) }
				on_shutdown.call
			rescue Exception => exc
				puts exc.message
				puts exc.backtrace
				terminate_gracefully
				on_shutdown.call 
			end
		}

	end

	def stop
		puts "calling stop...."
		# terminate_gracefully
		# @shutdown = true
	end

	#OVERRIDES START - more like monkey patches
  	def spawn_processes
	  	puts "override spawn process"
	    @processes.each do |process|
	    	b = @registered_bundles[process]
	    	1.upto(b.status.scale) do |n|
	    		pid =  spawn_process(b,process,n)  
	    		b.status.add_pid(pid)
	    	end
	    	# update_bundle_running_status(b)
		end
	    pp @running
	end

	def register(name, command, options={})
	  	options[:env] ||= env
	  	options[:cwd] ||= File.dirname(command.split(" ").first)
	  	process = Foreman::Process.new(command, options)
	  	@names[process] = name
	  	@processes << process
	  	process #just return the process!
	end

	def handle_term_signal
	    puts "custom SIGTERM received"
	    terminate_gracefully
	    @shutdown = true
	    sleep(1)
	    # exit(1)
	    # Process.kill("EXIT", 0)
	end

	def handle_interrupt
	    puts "custom SIGINT received"
	    terminate_gracefully
	    @shutdown = true
	    sleep(1)
	    # exit(1)	    
	    # Process.kill("EXIT", 0)
	end

	def handle_hangup
	    puts "custom SIGHUP received"
	    terminate_gracefully
	    @shutdown = true
	    sleep(1)
	    # exit(1)
	    # Process.kill("EXIT", 0)
	end

	#OVERRIDES END

	def name_for_bundle_index(bundle,index)
		[ bundle.id, index.to_s ].compact.join(".")
	end

	def spawn_process(bundle,process,n)
	  # 1.upto(count) do |n|
        reader, writer = create_pipe
        begin
           	puts "starting process"
           	name = name_for_bundle_index(bundle, n)
           	pid = process.run(:output => writer, :env => {
            	"PORT" => port_for(process, n).to_s,
            	"PS" => name
           	})
          	puts "started process"
          	@process_ids[name] = pid
         	writer.puts "started with pid #{pid}"
        rescue Errno::ENOENT
          writer.puts "unknown command: #{process.command}"
        end
        @running[pid] = [process, n]
        @readers[pid] = reader
      # end
      	pid
	end

	def kill_process_and_wait(pid,signal="SIGTERM")
	      begin
	      	puts "killing process #{pid}"
	        Process.kill signal, pid
	      rescue Errno::ESRCH, Errno::EPERM
	      end
	      begin
	      	Process.wait(pid)
	      	puts "process finished #{pid}"
		  rescue Errno::ECHILD
		  end
	end

	def bundle(id)
		@bundles.select{ |b| b.id == id }.first
	end

	def scale(name,count)
		bundle_to_scale = bundle(name)
		
		apply_scale(bundle_to_scale,count)

		bundle_to_scale.status.scale = count

		flush
	end

	def apply_scale(bundle,count)
		pp bundle
		puts bundle.status.scale.class
		puts "currently running #{bundle.status.scale}. Wanted #{count}"
		if (bundle.status.scale > count)
			# downscale to count. kill processes
			bundle.status.scale.downto(count+1) do |n|
				puts "apply_scale down #{n}"
				if n >= 1
					remove_process(bundle,n)
					sleep(0.1)
				end
			end
		elsif (bundle.status.scale < count)
			# add processes
			# diff = count - bundle.status.scale
			puts "adding process."
			(bundle.status.scale+1).upto(count) do |n|
				puts "apply_scale up #{n}" 
				if n >= 1
					add_process(bundle,n)
					sleep(0.1)
				end
			end
		end
		# update_bundle_running_status(bundle)
	end

	def add_process(bundle,index)
		proc = process(bundle.id)
    	pid =spawn_process(bundle,proc,index)  
    	bundle.status.add_pid(pid)
    	# update_bundle_running_status(b)
	end

	def remove_process(bundle,index)
		name = name_for_bundle_index(bundle, index)
		pid = @process_ids[name]
		unless pid.nil?
			kill_process_and_wait(pid)
			# bundle.scale.pids.reject! { |p| p == pid }
			bundle.status.remove_pid(pid)
		end
	end

	# def stop_all

	# end

	# def stop_plug

	# end

	# def start_plug

	# end

	def process_count
		@process_ids.count
	end

	def process_ids
		@process_ids
	end

	def bundles_json
		begin
			FileUtils.mkdir_p File.dirname(BUNDLES_JSON)
			File.read(BUNDLES_JSON)
		rescue Exception => exc
			# puts exc.message
			# puts exc.backtrace
			"[]"
		end
	end

	def bundles_status_json
		begin
			FileUtils.mkdir_p File.dirname(BUNDLES_STATUS_JSON)
			File.read(BUNDLES_STATUS_JSON)
		rescue Exception => exc
			# puts exc.message
			# puts exc.backtrace
			"[]"
		end
	end

	def load_bundles
		bundles = JSON.parse(bundles_json).map { |b| Ignition::Bundle.new(b) }
		# pp @bundles 

		JSON.parse(bundles_status_json).each do |bs| 
			# bundles
			b = bundles.select{ |b| b.id == bs.id }.first 
			unless b.nil?
				b.status = Ignition::BundleStatus.new(bs.status)
			end
		end
		remove_old_pids(bundles)
	end

	def flush
		remove_old_pids(@bundles)
		File.write(BUNDLES_JSON,JSON.pretty_generate( @bundles.map{|b| b.reject{|k| k == "status"}} ))
		File.write(BUNDLES_STATUS_JSON,JSON.pretty_generate( @bundles.map do |b| 
				{ "id" => b.id , "status" => b.status } 
		end
		))
	end

	def remove_old_pids(bundles)
		bundles.each do |b|
			b.status.clean_pids
		end
	end

	# 
	def new_message(message)
		# pp message

		@network_daemons[message.handle] = Time.now
	end

	def network_daemons
		@network_daemons
	end

	def start_network_ping
		Thread.new {
			begin
				loop {
					sleep(1)
					# puts "sending ping"
					@broadcast_client.transmit("PING")
				}
	      rescue Exception => exc
	        puts exc.message
	        puts exc.backtrace
	      end
		}
	end

	def recent_logs
		@recent_logs
	end

	def logs(bundle_id)
		# @recent_logs[bundle_id]
		proc_ids = @recent_logs.keys.select { |a| a.scan(/#{bundle_id}.\d+/).size > 0 }
	
		proc_ids.map { |b| @recent_logs[b].to_a }.flatten
	end
end